#!/usr/bin/python
import numpy as np
import math
import matplotlib
from matplotlib import pyplot as plt
import matplotlib.cm as cm
import skrf as rf

#Code to calculate the normalized errors expected from a multiline TRL VNA calibration

def calcMicrostripPropCon(epsdiel,dielthick,tand,linewidth,metalcond,Z0,freq):
	eps0 = 8.85419*1e-12
	mu0 = 4*math.pi*1e-7
	#numerical data for silicon microstrip line, comment out for other types
	epsdiel = 11.9
	dielthick = 250*1e-6
	tand = 0.001
	linewidth = 204*1e-6
	metalcond = 4.1*1e7
	Z0 = 50
	epseff = (epsdiel+1)/2 + (epsdiel-1)/(2*np.sqrt(1+12*dielthick/linewidth))
	#print(epseff)
	if epsdiel > 1:
		alphad = 2*math.pi*freq*1e9*np.sqrt(eps0*mu0)*epsdiel*(epseff-1)*tand/(2*np.sqrt(epseff)*(epsdiel-1))
	else:
		alphad = 0
	#print(alphad)
	alphacond = np.sqrt(2*math.pi*freq*1e9*mu0/(2*metalcond))/(Z0*linewidth)
	#print(alphacond)
	beta = 2*math.pi*freq*1e9*np.sqrt(mu0*eps0*epseff)
	#print(beta)
	return alphad+alphacond+1j*beta
	
def getCPWPropCons(freqs,w,s,ep_r,t,rho):
	freq = rf.Frequency(freqs[0],freqs[len(freqs)-1],len(freqs),'ghz')
	cpw = rf.media.CPW(freq, w, s, ep_r, t, rho)
	propcons = cpw.gamma()
	return propcons
	
def calcPropConError(lines,freq):
	linedelts = []
	for i in range(1,len(lines)):
		linedelts.append(lines[0]-lines[i])
	covmat = np.identity(len(lines)-1) - 1/(len(lines))*np.ones((len(lines)-1,len(lines)-1))
	linedelts = np.array(linedelts)
	sigma2 = np.dot(np.transpose(linedelts),np.dot(covmat,linedelts))
	#print(linedelts)
	#print(covmat)
	#print(np.dot(covmat,linedelts))
	return 1/np.sqrt(sigma2)
	
def calcCalConError(lines,freq,propcon):
	covmata = []
	covmatb = []
	for i in range(0,len(lines)-1):
		covmatarow = []
		covmatbrow = []
		for j in range(0,len(lines)-1):
			E10 = np.exp(-propcon*lines[0])
			E20 = np.exp(propcon*lines[0])
			E1I = np.exp(-propcon*lines[i+1])
			E2I = np.exp(propcon*lines[i+1])
			E1J = np.exp(-propcon*lines[j+1])
			E2J = np.exp(propcon*lines[j+1])
			E10I = np.exp(-propcon*(lines[i+1]-lines[0]))
			E20I = np.exp(propcon*(lines[i+1]-lines[0]))
			E10J = np.exp(-propcon*(lines[j+1]-lines[0]))
			E20J = np.exp(propcon*(lines[j+1]-lines[0]))
			if i == j:
				kdelt = 1
			else:
				kdelt = 0
			covmatarow.append((np.conj(E10I)*E10J+kdelt*np.absolute(E20I)**2+(1+kdelt)*np.absolute(E10)**2*np.conj(E1I)*E1J)/(np.conj(E20I-E10I)*(E20J-E10J)))
			covmatbrow.append((np.conj(E20I)*E20J+kdelt*np.absolute(E10I)**2+(1+kdelt)*np.absolute(E20)**2*np.conj(E2I)*E2J)/(np.conj(E20I-E10I)*(E20J-E10J)))
		covmata.append(covmatarow)
		covmatb.append(covmatbrow)
	covmatainv = np.linalg.inv(covmata)
	covmatbinv = np.linalg.inv(covmatb)
	suma = np.dot(np.transpose(np.ones(len(lines)-1)),np.dot(covmatainv,np.ones(len(lines)-1)))
	sumb = np.dot(np.transpose(np.ones(len(lines)-1)),np.dot(covmatbinv,np.ones(len(lines)-1)))
	#print(suma)
	#print(sumb)
	return 1/np.sqrt(np.real(suma)),1/np.sqrt(np.real(sumb))
	
def calcEffPhaseDiff(propcon,lij):
	E1IJ = np.exp(-propcon*lij)
	E2IJ = np.exp(propcon*lij)
	diff = np.absolute(0.5*(E2IJ-E1IJ))
	if diff >= 1:
		return 90
	else:
		phieff = 180/math.pi*math.asin(diff)
		#if phieff < 0:
		#	phieff = 180 + phieff
		return phieff

def getEffPhases(freqs,propcons,lines):
	linedelts = []
	for i in range(0,len(lines)):
		for j in range(0,i):
			linedelts.append(lines[j]-lines[i])
	phases = []
	for lij in linedelts:
		phaseline = []
		for i in range(0,len(freqs)):
			phaseline.append(calcEffPhaseDiff(propcons[i],lij))
		phases.append(phaseline)
	phases = np.array(phases)
	return phases

	
def getCalConErrors(freqs,lines,propcons):
	alphasigma = []
	betasigma = []
	for i in range(0,len(freqs)):
		(sigmaa,sigmab) = calcCalConError(lines,freqs[i],propcons[i])
		alphasigma.append(sigmaa)
		betasigma.append(sigmab)
	
	alphasigma = np.array(alphasigma)
	betasigma = np.array(betasigma)
	return alphasigma,betasigma
	
def getAllCalConErrors(freqs,lines,propcons):
	linetemps = []
	for i in range(0,len(lines)):
		for j in range(0,i):
			linerow = []
			linerow.append(lines[i])
			linerow.append(lines[j])
			linetemps.append(linerow)
	linetemps = np.array(linetemps)
	alphasigmas = []
	betasigmas = []
	for linetemp in linetemps:
		(alpsig,betasig) = getCalConErrors(freqs,linetemp,propcons)
		alphasigmas.append(alpsig)
		betasigmas.append(betasig)
	alphasigmas = np.array(alphasigmas)
	betasigmas = np.array(betasigmas)
	return alphasigmas,betasigmas
	
def getAveVariance(alphasigma,betasigma):
	alphavar = 0
	betavar = 0
	for i in range(0,len(alphasigma)):
		alphavar = alphavar + alphasigma[i]**2/len(alphasigma)
	for i in range(0,len(betasigma)):
		betavar = betavar + betasigma[i]**2/len(betasigma)
	return alphavar,betavar
	
def getTRLCostFunction(lines,freqs,temp):
	lines = np.array(lines)#lines in mm
	lines = lines/1000 #lines in m
	if temp == 4:
		propcons = getCPWPropCons(freqs,w=24e-6, s=13e-6, ep_r=11.42, t=200e-9, rho=0.022e-8) #4K
	elif temp == 300:
		propcons = getCPWPropCons(freqs,w=24e-6, s=13e-6, ep_r=11.9, t=200e-9, rho=2.27e-8) #300K
	else:
		print("Pick 4K or 300K")
	(alphasigma,betasigma) = getCalConErrors(freqs,lines,propcons)
	phases = getEffPhases(freqs,propcons,lines)
	(alphavar,betavar) = getAveVariance(alphasigma,betasigma)
	return 1/alphavar + 1/betavar
